#include "fastlz.h"
#include "filesystem.h"
#include "nds/arm9/console.h"
#include "raylibds.hpp"
#include <algorithm>
#include <nds.h>
#include <maxmod9.h>

#define PALETTE_SIZE 256
#define FRAME_SIZE 256*192
#define CHUNK_SIZE 10
#define QUEUE_SIZE 6
#define FRAMERATE 30

#define STREAM_CHUNK_SIZE 32000
#define STREAM_QUEUE_SIZE 4
#define MMSTREAM_BUF_SIZE 9600

void wait_forever(const char* msg);
void LoadNextChunk(void *output);
void TimerCallback();
void VBLCallback();
void FrameStep();
int ThreadEntrypoint(void *arg);
mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format );

class CircularQueue {
public:
  int front, rear, size, current;
  CircularQueue(int s) {
    size = s;
    front = rear = -1;
		current = 0;
  }

  bool isFull() {
    return (front == 0 && rear == size - 1) || (rear == (front - 1) % (size - 1));
  }

  bool isEmpty() {
    return front == -1;
  }

  void push() {
    if (isFull()) {
      fprintf(stderr, "Queue is full. Cannot push\n");
      return;
    }

    if (isEmpty()) {
      front = rear = 0;
    } else {
      rear = (rear + 1) % size;
    }

		current++;
  }

  void pop() {
    if (isEmpty()) {
      fprintf(stderr, "Queue is empty. Cannot pop\n");
      return;
    }

    if (front == rear) {
      front = rear = -1;
    } else {
      front = (front + 1) % size;
    }

		current--;
  }
};

int ptr_background;
int ptr_subbackground;
FILE *file_image = nullptr;
FILE *file_music = nullptr;

int chunk_current = 0;

uint8_t *frame_decompress;
uint8_t *frame_buffer;
uint8_t *stream_buffer;

int stream_index;
int stream_buffer_current;
bool stream_need_update = false;

CircularQueue frame_queue(QUEUE_SIZE);
CircularQueue stream_queue(STREAM_QUEUE_SIZE);

int music_buffer_size = 4000*16;

int main(void) {

	videoSetMode(MODE_5_3D);
	videoSetModeSub(MODE_5_2D);

	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankC(VRAM_C_SUB_BG);
	vramSetBankF(VRAM_F_TEX_PALETTE);

	// consoleDemoInit();
	consoleDebugInit(DebugDevice_NOCASH);

	frame_decompress = (uint8_t*)malloc(FRAME_SIZE*CHUNK_SIZE);
	frame_buffer = (uint8_t*)malloc(FRAME_SIZE*CHUNK_SIZE*QUEUE_SIZE);
	stream_buffer = (uint8_t*)malloc(STREAM_CHUNK_SIZE*STREAM_QUEUE_SIZE);

	sassert(
		frame_buffer != nullptr ||
		frame_decompress != nullptr ||
		stream_buffer != nullptr,
		"failed to allocate memmory");

	bool nitrofs = nitroFSInit(NULL);
	sassert(nitrofs, "error nitrofs");

	file_image = fopen("nitro:/image.bin", "rb");
	sassert(file_image != nullptr, "failed to load image.bin");

	file_music = fopen("nitro:/music.raw", "rb");
	sassert(file_image != nullptr, "failed to load music.raw");

	int pal_len;
	unsigned char* pal = Raylib::LoadFile("nitro:/palette.bin", pal_len);

	ptr_background = bgInit(3, BgType_Bmp8, BgSize_B8_256x256, 0, 0);
	ptr_subbackground = bgInitSub(3, BgType_Bmp8, BgSize_B8_256x256, VRAM_C_SUB_BG, 0);

	DC_FlushRange(pal, pal_len);
	dmaCopy(pal, BG_PALETTE, pal_len);
	dmaCopy(pal, BG_PALETTE_SUB, pal_len);

	free(pal);

	while(!frame_queue.isFull()) {
		frame_queue.push();
		LoadNextChunk(&frame_buffer[FRAME_SIZE*CHUNK_SIZE*frame_queue.rear]);
	}

	while(!stream_queue.isFull()) {
		stream_queue.push();
		fread(&stream_buffer[STREAM_CHUNK_SIZE*stream_queue.rear], sizeof(uint8_t), STREAM_CHUNK_SIZE, file_music);
	}
	stream_index = 0;

	DC_FlushAll();

	mm_ds_system sys;
	sys.mod_count 			= 0;
	sys.samp_count			= 0;
	sys.mem_bank			  = 0;
	sys.fifo_channel		= FIFO_MAXMOD;
	mmInit( &sys );

	mm_stream mystream;
	mystream.sampling_rate	= 22050;
	mystream.buffer_length	= MMSTREAM_BUF_SIZE;
	mystream.callback		= on_stream_request;
	mystream.format			= MM_STREAM_16BIT_MONO;
	mystream.timer			= MM_TIMER2;
	mystream.manual			= false;

	timerStart(0, ClockDivider_1024, TIMER_FREQ_1024(FRAMERATE), TimerCallback);
	cothread_create(ThreadEntrypoint, NULL, 0, COTHREAD_DETACHED);

	mmStreamOpen( &mystream );

	while(1) {
		cothread_yield_irq(IRQ_VBLANK);
		scanKeys();

		if (keysUp() & KEY_A) break;
	}

	return 0;
}

int ThreadEntrypoint(void *arg)
{
	while(1) {
		if (!frame_queue.isFull())
		{
			frame_queue.push();
			LoadNextChunk(&frame_buffer[FRAME_SIZE*CHUNK_SIZE*frame_queue.rear]);
		}
		cothread_yield();

		if (!stream_queue.isFull())
		{
			stream_queue.push();
			
			fread(stream_buffer+ stream_queue.rear*STREAM_CHUNK_SIZE, sizeof(uint8_t), STREAM_CHUNK_SIZE, file_music);
			if (feof(file_music)) {
				memset(stream_buffer+ stream_queue.rear*STREAM_CHUNK_SIZE, 0, STREAM_CHUNK_SIZE);
				mmStreamClose();
			}
		}
		cothread_yield();
	}
	
	return 0;
}

void FrameStep()
{
	// if (feof(file_image)) return;

	dmaCopyAsynch(&frame_buffer[FRAME_SIZE*CHUNK_SIZE*frame_queue.front + FRAME_SIZE*chunk_current], bgGetGfxPtr(ptr_background), FRAME_SIZE);
	dmaCopyAsynch(&frame_buffer[FRAME_SIZE*CHUNK_SIZE*frame_queue.front + FRAME_SIZE*chunk_current], bgGetGfxPtr(ptr_subbackground), FRAME_SIZE);

	chunk_current++;

	if (chunk_current >= CHUNK_SIZE) {
		chunk_current = 0;
		frame_queue.pop();
	}
}

void TimerCallback()
{
	FrameStep();
}

void LoadNextChunk(void *output)
{
	uint32_t compress_size;
	fread(&compress_size, sizeof(uint32_t), 1, file_image);

	fread(frame_decompress, sizeof(uint8_t), compress_size, file_image);
	fastlz_decompress(frame_decompress, compress_size, output, FRAME_SIZE*CHUNK_SIZE);
}

mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format ) {
	
	size_t samplesize = 1;
	switch(format){
		case MM_STREAM_8BIT_MONO: samplesize = 1; break;
		case MM_STREAM_8BIT_STEREO: samplesize = 2; break;
		case MM_STREAM_16BIT_MONO: samplesize = 2; break;
		case MM_STREAM_16BIT_STEREO: samplesize = 4; break;
	}
	int len = length*samplesize;

	if (feof(file_music))
	{
		mmStreamClose();
		uint8_t temp[len];
		memset(temp, 0, len);
		DC_FlushAll();
		dmaCopyAsynch(temp, dest, len);
		return length;
	}

	int bytesToCopy = std::min(len, STREAM_CHUNK_SIZE-stream_index);
	dmaCopyAsynch(stream_buffer + stream_queue.front*STREAM_CHUNK_SIZE + stream_index, dest, bytesToCopy);
	stream_index += bytesToCopy;

	if (stream_index >= STREAM_CHUNK_SIZE)
	{
		stream_queue.pop();
		stream_index = len - bytesToCopy;
		// stream_index = 0;

		dmaCopyAsynch(stream_buffer + stream_queue.front*STREAM_CHUNK_SIZE, dest+bytesToCopy, stream_index);
	}

	// return bytesToCopy/2;
	return length;
}